# mongodb.py
from abstract_source import AbstractSource
import copy

from ..conf import merge_conf
import urlparse

from pymongo import MongoClient

connections = {}


class Mongodb(AbstractSource):
    @classmethod
    def from_settings(self, settings):
        if not 'mongodb' in settings or not 'collection' in settings or \
                settings['mongodb'] == '' or settings['collection'] == '':
            raise Exception(
                "Erroneous mongodb settings, "
                "needs a collection and mongodb setting",
                settings)

        cx_uri = urlparse.urlsplit(settings["mongodb"])
        db_name = cx_uri.path[1:]
        if db_name == "":
            raise Exception(
                "Erroneous mongodb settings, "
                "missing db_name", settings)

        cx_uri = urlparse.urlunsplit(
            (cx_uri.scheme, cx_uri.netloc, "/", cx_uri.query, cx_uri.fragment))
        options = copy.deepcopy(settings)
        del options['mongodb']
        del options['collection']

        return Mongodb(
            self.connection_for_uri(cx_uri),
            db_name, settings['collection'], options)

    @classmethod
    def connection_for_uri(self, uri):
        if not uri in connections:
            connections[uri] = MongoClient(uri)
        return connections[uri]

    TOP_LEVEL_COLLECTIONS = ['fwissr']

    def __init__(self, conn, db_name, collection_name, options):
        super(Mongodb, self).__init__(options)

        self._conn = conn
        self._db_name = db_name
        self._collection_name = collection_name

        self._collection = None

    def db_name():
        doc = "The db_name property."

        def fget(self):
            return self._db_name

        def fset(self, value):
            self._db_name = value

        def fdel(self):
            del self._db_name
        return locals()
    db_name = property(**db_name())

    def collection_name():
        doc = "The collection_name property."

        def fget(self):
            return self._collection_name

        def fset(self, value):
            self._collection_name = value

        def fdel(self):
            del self._collection_name
        return locals()
    collection_name = property(**collection_name())

    def fetch_conf(self):
        result = {}
        result_part = result

        if not self._collection_name in Mongodb.TOP_LEVEL_COLLECTIONS and (
                not 'top_level' in self._options
                or not self._options["top_level"]):
            for key_part in self._collection_name.split("."):
                if not key_part in result_part:
                    result_part[key_part] = {}
                result_part = result_part[key_part]

        conf = {}
        for doc in self.collection().find():
            key = doc['_id']
            if not 'value' in doc:
                del doc['_id']
                value = doc
            else:
                value = doc['value']
            conf[key] = value

        merge_conf(result_part, conf)
        return result

    def collection(self):
        if self._collection is None:
            self._collection = self._conn[self._db_name][self._collection_name]
        return self._collection



# file.py
from abstract_source import AbstractSource
import copy
import glob
import os

from ..conf import parse_conf_file, merge_conf


class File(AbstractSource):
    TOP_LEVEL_CONF_FILES = ['fwissr']

    @classmethod
    def from_path(self, path, options={}):
        if path is None or path == '':
            raise Exception("Unexpected file path", path)
        return File(path, options)

    @classmethod
    def from_settings(self, settings={}):
        mine = copy.deepcopy(settings)
        del mine['filepath']
        return self.from_path(settings['filepath'], mine)

    def __init__(self, path, options={}):
        super(File, self).__init__(options)
        if not os.path.exists(path):
            raise Exception("Missing file", path)
        self._path = path

    def path():
        doc = "The path property."

        def fget(self):
            return self._path

        def fset(self, value):
            self._path = value

        def fdel(self):
            del self._path
        return locals()
    path = property(**path())

    def fetch_conf(self):
        result = {}

        if os.path.isdir(self._path):
            conf_files = []
            file_types = ("*.json", "*.yml", "*.yaml")
            for type in file_types:
                conf_files.extend(glob.glob(os.path.join(self._path,  type)))
            conf_files.sort()
        else:
            conf_files = [self._path]

        for conf in conf_files:
            if os.path.isfile(conf):
                self.merge_conf_file(result, conf)

        return result

    def merge_conf_file(self, result, conf_file_path):
        conf = parse_conf_file(conf_file_path)
        conf_file_name = os.path.splitext(os.path.basename(conf_file_path))[0]
        result_part = result

        if not conf_file_name in File.TOP_LEVEL_CONF_FILES \
            and (
                not "top_level" in self._options
                or not self._options["top_level"]):
            for key_part in conf_file_name.split('.'):
                if not key_part in result_part:
                    result_part[key_part] = {}
                result_part = result_part[key_part]

        return merge_conf(result_part, conf)



from threading import RLock, Thread

from conf import merge_conf
import UserDict
import UserList
import time


class ReadOnlyList(UserList.UserList):
    def __setitem__(self, i, item):
        raise TypeError

    def update(self, array=None):
        if array is None:
            pass
        elif isinstance(array, UserList.UserList):
            self.data = array.data
        elif isinstance(array, type([])):
            self.data = array
        else:
            raise TypeError


class ReadOnlyDict(UserDict.IterableUserDict):
    def __setitem__(self, key, item):
        raise TypeError

    def __delitem__(self, key):
        raise TypeError

    def clear(self):
        raise TypeError

    def pop(self, key, *args):
        raise TypeError

    def popitem(self):
        raise TypeError

    def update(self, dict=None):
        if dict is None:
            pass
        elif isinstance(dict, UserDict.UserDict):
            self.data = dict.data
        elif isinstance(dict, type({})):
            self.data = dict
        else:
            raise TypeError


class ReloadThread(Thread):
    def __init__(self, registry):
        super(ReloadThread, self).__init__()
        self._registry = registry

    def run(self):
        count = 0
        while True:
            time.sleep(1)
            if count == self._registry.refresh_period:
                self._registry.load()
                count = 0
            count += 1


class Registry:
    DEFAULT_REFRESH_PERIOD = 30

    def __init__(self, refresh_period=DEFAULT_REFRESH_PERIOD):
        self._refresh_period = refresh_period

        self._registry = {}
        self._frozen = False
        self.sources = []

        self.semaphore = RLock()
        self.refresh_thread = None

    def frozen():
        doc = "The frozen property."

        def fget(self):
            return self._frozen

        def fset(self, value):
            self._frozen = value

        def fdel(self):
            del self._frozen

        return locals()
    frozen = property(**frozen())

    def refresh_period():
        doc = "The refresh_period property."

        def fget(self):
            return self._refresh_period
        return locals()
    refresh_period = property(**refresh_period())

    def add_source(self, source):
        with self.semaphore:
            self.sources.append(source)

        if self._frozen:
            self.reload()
        else:
            with self.semaphore:
                merge_conf(self._registry, source.get_conf())

        self.ensure_refresh_thread()
        self.ensure_frozen()

    def reload(self):
        self.reset()
        self.load()

    def deep_freeze(self, what):
        if isinstance(what, dict) or isinstance(what, UserDict.UserDict):
            w = {}
            for (key, value) in what.items():
                if(isinstance(value, dict)
                        or isinstance(value, UserDict.UserDict)):
                    w[key] = self.deep_freeze(value)
                elif (isinstance(value, type([]))
                        or isinstance(value, UserList.UserList)):
                    w[key] = self.deep_freeze(value)
                else:
                    w[key] = value
            r = ReadOnlyDict()
            r.update(w)
            return r
        elif isinstance(what, type([])) or isinstance(what, UserList.UserList):
            r = ReadOnlyList()
            r.update(what)
            return r
        else:
            raise Exception("Deep Freezine failed for", what)

    def get(self, key):
        key_ary = key.split("/")

        if key_ary[0] == "":
            key_ary.pop(0)

        if key_ary[-1] == "":
            key_ary.pop(len(key_ary)-1)

        cur_hash = self._registry
        for key_component in key_ary:
            if key_component in cur_hash:
                cur_hash = cur_hash[key_component]
            else:
                return None

        return cur_hash

    def __getitem__(self, key):
        return self.get(key)

    def keys(self):
        result = []
        self._keys(result, [], self._registry)
        return sorted(result)

    def dump(self):
        return self._registry

    def have_refreshable_source(self):
        with self.semaphore:
            return True in [source.can_refresh() for source in self.sources]
            return False

    def ensure_refresh_thread(self):
        if(self.refresh_period > 0) and self.have_refreshable_source() \
            and (
                self.refresh_thread is None or
                not self.refresh_thread.is_alive()):
            # re-start refresh thread
            self.refresh_thread = ReloadThread(self)
            self.refresh_thread.daemon = True
            self.refresh_thread.start()

    def ensure_frozen(self):
        if not self._frozen:
            with self.semaphore:
                self._registry = self.deep_freeze(self._registry)
                self._frozen = True

    def reset(self):
        with self.semaphore:
            self._registry = {}
            self._frozen = False
            [source.reset() for source in self.sources]

    def load(self):
        with self.semaphore:
            self._registry = {}
            self.frozen = False
            for source in self.sources:
                source_conf = source.get_conf()
                merge_conf(self._registry, source_conf)
        # print "Reloaded with content: %s" % self

    def registry():
        doc = "The registry property."

        def fget(self):
            self.ensure_refresh_thread
            return self._registry
        return locals()
    registry = property(**registry())

    def _keys(self, result, key_ary, dct):
        for (key, value) in dct.items():
            key_ary.append(key)
            result.append("/%s" % ("/".join(key_ary)))
            if isinstance(value, dict) or isinstance(value, UserDict.UserDict):
                self._keys(result, key_ary, value)
            key_ary.pop()



from file import File
from mongodb import Mongodb


class Source(object):
    @classmethod
    def from_settings(self, settings):
        if not isinstance(settings, dict):
            raise Exception("Unexpected source settings class", settings)

        if 'filepath' in settings:
            return File.from_settings(settings)
        elif 'mongodb' in settings:
            return Mongodb.from_settings(settings)
        else:
            raise Exception("Unexpected source settings kind", settings)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
from registry import Registry
from source.source import Source
from conf import parse_conf_file, merge_conf


class FwissrModule(object):
    # main conf file
    MAIN_CONF_FILE = "fwissr.json"
    # default path where main conf file is located
    DEFAULT_MAIN_CONF_PATH = "/etc/fwissr"
    # default directory (relative to current user's home)
    # where user's main conf file is located
    DEFAULT_MAIN_USER_CONF_DIR = ".fwissr"

    def __init__(self):
        self._main_conf_path = self.DEFAULT_MAIN_CONF_PATH
        self._main_user_conf_path = os.path.join(
            self.find_home(),
            self.DEFAULT_MAIN_USER_CONF_DIR)
        self._global_registry = None
        self._main_conf = None
        self._main_conf_file = None
        self._main_user_conf_file = None

    def main_conf_path():
        doc = "The main_conf_path property."

        def fget(self):
            return self._main_conf_path

        def fset(self, value):
            self._main_conf_path = value
        return locals()
    main_conf_path = property(**main_conf_path())

    def main_user_conf_path():
        doc = "The main_user_conf_path property."

        def fget(self):
            return self._main_user_conf_path

        def fset(self, value):
            self._main_user_conf_path = value
        return locals()
    main_user_conf_path = property(**main_user_conf_path())

    def find_home(self):
        for v in ('HOME', 'USERPROFILE'):
            if v in os.environ:
                return os.environ[v]

        if "HOMEDRIVE" in os.environ and "HOMEPATH" in os.environ:
            return "%s:%s" % (os.environ["HOMEDRIVE"], os.environ["HOMEPATH"])

        try:
            return os.path.expanduser("~")
        except Exception:
            if os.sep == '\\':
                return "C:/"
            else:
                return "/"

    def parse_args(self, argv):
        pass

    def global_registry(self):
        """Global Registry


        NOTE: Parses main conf files (/etc/fwissr/fwissr.json
            and ~/.fwissr/fwissr.json)

        then uses 'fwissr_sources' setting to setup additional sources

        Example of /etc/fwissr/fwissr.json file:

         {
           'fwissr_sources': [
             { 'filepath': '/mnt/my_app/conf/' },
             { 'filepath': '/etc/my_app.json' },
             { 'mongodb': 'mongodb://db1.example.net/my_app',
                 'collection': 'config', 'refresh': true },
           ],
           'fwissr_refresh_period': 30,
        }
        access global registry with Fwissr['/foo/bar']
        """
        if self._global_registry is None:
            if 'fwissr_refresh_period' in self.main_conf:
                result = Registry(
                    refresh_period=self.main_conf['fwissr_refresh_period'])
            else:
                result = Registry()

            if (os.path.exists(self.main_conf_path) or
                os.path.exists(self.main_user_conf_path)):

                if os.path.exists(self.main_conf_path):
                    result.add_source(
                        Source.from_settings(
                            {'filepath': self.main_conf_path}))

                if os.path.exists(self.main_user_conf_path):
                    result.add_source(
                        Source.from_settings(
                            {'filepath': self.main_user_conf_path}))

                if 'fwissr_sources' in self.main_conf:
                    for source in self.main_conf['fwissr_sources']:
                        result.add_source(Source.from_settings(source))

            self._global_registry = result
        return self._global_registry

    def main_conf():
        doc = "The main_conf property."

        def fget(self):
            if self._main_conf is None:
                result = {}
                if os.path.exists(self.main_conf_file):
                    result = merge_conf(
                        result,
                        parse_conf_file(self.main_conf_file))

                if os.path.exists(self.main_user_conf_file):
                    result = merge_conf(
                        result,
                        parse_conf_file(self.main_user_conf_file))

                self._main_conf = result
            return self._main_conf
        return locals()
    main_conf = property(**main_conf())

    def main_conf_file():
        doc = "The main_conf_file property."

        def fget(self):
            if self._main_conf_file is None:
                self._main_conf_file = os.path.join(
                    self.main_conf_path,
                    self.MAIN_CONF_FILE)

            return self._main_conf_file
        return locals()
    main_conf_file = property(**main_conf_file())

    def main_user_conf_file():
        doc = "The main_user_conf_file property."

        def fget(self):
            if self._main_user_conf_file is None:
                self._main_user_conf_file = os.path.join(
                    self.main_user_conf_path,
                    self.MAIN_CONF_FILE)

            return self._main_user_conf_file
        return locals()
    main_user_conf_file = property(**main_user_conf_file())

    def get(self, key):
        return self.global_registry().get(key)

    def __getitem__(self, key):
        return self.global_registry().get(key)

    def keys(self):
        return self.global_registry().keys()

    def dump(self):
        return self.global_registry().dump()

Fwissr = FwissrModule()



# conf.py
import os
import json
import yaml


def merge_conf(to_hash, other_hash, path=[]):
    "merges other_hash into to_hash"
    for key in other_hash:
        if (key in to_hash and isinstance(to_hash[key], dict)
                and isinstance(other_hash[key], dict)):
            merge_conf(to_hash[key], other_hash[key], path + [str(key)])
        else:
            to_hash[key] = other_hash[key]
    return to_hash


def parse_conf_file(conf_file_path):

    ext = os.path.splitext(conf_file_path)[1]
    if ext == ".json":
        return json.load(open(conf_file_path))
    elif ext == ".yaml" or ext == ".yml":
        return yaml.load(open(conf_file_path))
    else:
        raise Exception("Unsupported conf file kind", conf_file_path)



class AbstractSource(object):

    def __init__(self, options={}):
        self.reset()
        self._options = options

    def reset(self):
        self._conf = None

    def can_refresh(self):
        if 'refresh' in self._options and self._options['refresh']:
            return True
        return False

    def get_conf(self):
        if self._conf is None or self.can_refresh():
            self._conf = self.fetch_conf()
        return self._conf

    def fetch_conf(self):
        raise Exception("Not implemented")
